#include <stdio.h>
#include <gtk/gtk.h>

/* --- structures --- */

typedef struct {
  void (*func) (gchar *);
  GtkWidget *filesel;
} typFileSelectionData;

/* --- globals --- */

static char *sFilename = NULL;

/* --- function declarations --- */

void GetFilename (char *sTitle, void (*callback) (char *));
void FileOk (GtkWidget *w, gpointer data);
static void destroy (GtkWidget *widget, gpointer *data);

/* --- functions --- */

/*
 * GetFilename
 *
 * Show a dialog box with a title and if "ok" is slected
 * call the function with the name of the file.
 *
 * Use as GetFilename("Open", Loadfile);
 */
void GetFilename (char *sTitle, void (*callback) (char *))
{
  GtkWidget *filew = NULL;
  typFileSelectionData *data;
  //  GtkWidget *localentry;
  //GtkWidget *dialog;
  GtkWidget *localbutton;

  /* --- Create a File selection widget --- */
  filew = gtk_file_selection_new (sTitle);
  
  // dialog = GTK_FILE_SELECTION(filew)->fileop_dialog = gtk_dialog_new();

  //  localentry = gtk_entry_new();
  //gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),localentry);
  //localentry = gtk_button_new_with_label("hello");
  //gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),localentry);
  //gtk_widget_show_all(dialog);

  // localbutton = gtk_button_new_with_label("Hello");
  //gtk_container_add(GTK_CONTAINER(GTK_FILE_SELECTION(filew)->main_vbox),localbutton);
  //gtk_widget_show(localbutton);
  
  data = g_malloc (sizeof (typFileSelectionData));
  data->func = callback;
  data->filesel = filew;

  gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(filew));

  // ddc 7jun06 moved this to before OK, and gave sFilename  some memory!!! 
  if (sFilename){  
    /* --- Set the default filename --- */
    gtk_file_selection_set_filename (GTK_FILE_SELECTION (filew),
  				    sFilename);
  } else { //--ddc for cripes sake, give it some memory!
    sFilename = g_malloc(120);
    strcpy(sFilename,"default");
    gtk_file_selection_set_filename (GTK_FILE_SELECTION (filew),
  				    sFilename);
  }


  /* --- connect the "ok" button --- */
  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filew)->ok_button),
		      "clicked", (GtkSignalFunc) FileOk, data);
  
  /* --- connect the "cancel" button --- */
  gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION (filew)->cancel_button),
			     "clicked", (GtkSignalFunc) gtk_widget_destroy,
			     (gpointer) filew);
  //
  //  if (sFilename){  
  //  /* --- Set the default filename --- */
  //  gtk_file_selection_set_filename (GTK_FILE_SELECTION (filew),
  //				    sFilename);
  //} else { //--ddc for cripes sake, give it some memory!
  //  sFilename = g_malloc(120);
  // }

  
  /* --- show the dialogue box --- */
  gtk_widget_show_all (filew);

  /* --- grab the focus --- */
  gtk_grab_add(filew);
}

/* 
 * FileOk
 *
 * The "OK" button has been clicked.
 * Call the function (func) to do what is needed to the file
 */
void FileOk (GtkWidget *w, gpointer data)
{
  char *sTempFile;
  typFileSelectionData *localdata;
  GtkWidget *filew;

  localdata = (typFileSelectionData *) data;
  filew = localdata->filesel;

  /* --- Which file? --- */
  sTempFile = gtk_file_selection_get_filename( GTK_FILE_SELECTION (filew));

  /* --- Free old memory --- */
  if (sFilename) g_free (sFilename);

  /* --- Duplicate the string --- */
  sFilename = g_strdup (sTempFile);

  /* --- Call the Function that does the work --- */
  (*(localdata->func)) (sFilename);

  /* --- Free the sFilename --- */
  g_free (sFilename);
  sFilename = NULL;

  /* --- Close the dialog box --- */
  gtk_widget_destroy (filew);
}

/*
 * destroy
 *
 * Function to handle the destroying of the dialog box.  We must release the focus that we grabbed.
 */
static void destroy (GtkWidget *widget, gpointer *data)
{
  /* --- remove the focus --- */
  gtk_grab_remove (widget);
  g_free(data);
}


