/* output.c
 * by John Pavan
 *
 * Handles the text output of gnuscope
 */

#include <gtk/gtk.h>
#include "gnuscopefuncs.h"
#include "gnuscopeglobals.h"

//--ddc 23jun06 define close for messagetext
void CloseMessageTextWidget(GtkWidget *);

/* --- structures --- */

/* --- globals --- */
/* --- should be included in gnuscopeglobals.h --- */

/* --- function declarations --- */
/* --- should be included in gnuscopefuncs.h --- */

/* --- functions --- */

/*
 * WriteMainText
 *
 * Writes a string to the main text box
 */
void WriteMainText(const char *newtext)
{
  //--kp 3may06  if (maintext != NULL){
  if (GTK_IS_WIDGET(maintext)){
    gtk_text_freeze(GTK_TEXT(maintext));
    if (gtk_text_get_length(GTK_TEXT(maintext)) > 0) {
      gtk_text_set_point(GTK_TEXT(maintext),GTK_POS_TOP);
      gtk_text_set_point(GTK_TEXT(maintext),GTK_POS_LEFT);
    }
    gtk_text_insert(GTK_TEXT(maintext),NULL,NULL,NULL,newtext,strlen(newtext));
    gtk_text_thaw(GTK_TEXT(maintext));
    gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(maintext)->vadj),0);
  } else {
    printf(newtext);
  }
}

/*
 * WriteTwodText
 *
 * Writes a string to the twod text box
 */
void WriteTwodText(const char *newtext)
{
  //--kp 3may06  if (twodtext != NULL) {
  if (GTK_IS_WIDGET(twodtext)) {
    gtk_text_freeze(GTK_TEXT(twodtext));
    if (gtk_text_get_length(GTK_TEXT(twodtext)) > 0) {
      gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(twodtext)->vadj),0);
      gtk_text_set_point(GTK_TEXT(twodtext),GTK_POS_TOP);
      gtk_text_set_point(GTK_TEXT(twodtext),GTK_POS_LEFT);
    }
    gtk_text_insert(GTK_TEXT(twodtext),NULL,NULL,NULL,newtext,strlen(newtext));
    gtk_text_thaw(GTK_TEXT(twodtext));
  }
}

/*
 * WriteMessageText
 *
 * Writes to the message text window
 * if the message text window does not exist (messagetext == null)
 * then a new message text window is created
 * the message text window should be used for text which is useful for
 * more than one display type (such as file input/output and pgam sorting messages)
 */
void WriteMessageText(const char *newtext)
{
  //--kp 03may06  if (messagetext == NULL) CreateNewMessageTextWindow();
  if (!GTK_IS_WIDGET(messagetext)) CreateNewMessageTextWindow();
  g_return_if_fail(GTK_IS_WIDGET(messagetext));
  gtk_text_freeze(GTK_TEXT(messagetext));
  if (gtk_text_get_length(GTK_TEXT(messagetext)) > 0) {
    gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(messagetext)->vadj),0);
    gtk_text_set_point(GTK_TEXT(messagetext),GTK_POS_TOP);
    gtk_text_set_point(GTK_TEXT(messagetext),GTK_POS_LEFT);
  }
  gtk_text_insert(GTK_TEXT(messagetext),NULL,NULL,NULL,newtext,strlen(newtext));
  gtk_text_thaw(GTK_TEXT(messagetext));
}

/* CreateNewMessageTextWindow
 * 
 * generates a new text window and assigns the "message text" to that window
 */
void CreateNewMessageTextWindow()
{
  GtkWidget *table;
  GtkWidget *hscrollbar;
  GtkWidget *vscrollbar;
  GtkWidget *localwindow;

  /* --- local window --- */
  localwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_default_size(GTK_WINDOW(localwindow),400,100);
  gtk_window_set_title(GTK_WINDOW(localwindow),"Gnuscope Messages");
  
  //--ddc 23jun06
    gtk_signal_connect (GTK_OBJECT(localwindow),"destroy",
			GTK_SIGNAL_FUNC(CloseMessageTextWidget),NULL);

  /* --- table for text and scrollbars --- */
  table = gtk_table_new(3,3,FALSE);
  gtk_container_add(GTK_CONTAINER(localwindow),table);
  gtk_table_set_row_spacing(GTK_TABLE(table),0,2);
  gtk_table_set_col_spacing(GTK_TABLE(table),2,0);
  gtk_widget_show(table);

  /* --- message text --- */
  messagetext = gtk_text_new(NULL,NULL);
  gtk_text_set_editable(GTK_TEXT(messagetext),FALSE);
  gtk_table_attach(GTK_TABLE(table),messagetext,0,1,0,1,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		   GTK_EXPAND | GTK_SHRINK | GTK_FILL,0,0);
  gtk_widget_show(messagetext);

  /* --- scrollbars --- */
  hscrollbar = gtk_hscrollbar_new(GTK_TEXT(messagetext)->hadj);
  gtk_table_attach(GTK_TABLE(table),hscrollbar,0,1,1,2,
                   GTK_EXPAND | GTK_FILL | GTK_SHRINK, GTK_FILL,0,0);
  gtk_widget_show (hscrollbar);
  vscrollbar = gtk_vscrollbar_new(GTK_TEXT(messagetext)->vadj);
  gtk_table_attach(GTK_TABLE(table),vscrollbar,1,2,0,1,
                   GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0,0);
  gtk_widget_show (vscrollbar);

  gtk_widget_show(localwindow);
}


//--ddc 23jun06 add a handler for window destroy
void CloseMessageTextWidget(GtkWidget *widget){
  gtk_widget_destroy(widget);
  messagetext=NULL;
}
  
