/* File: printing.c
 *
 * by: John Pavan
 *
 * Printing support for gnuscope
 */

/* --- includes --- */

#include <gtk/gtk.h>
#include "gtkgraph.h"
#include "gtktwodplot.h"
#include <stdio.h>
//--ddc 3jun08 add missing function prototypes (stdlib,string,time,unistd)
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/types.h>

/* --- structs --- */

struct string_list {
  struct string_list *ptr;  // pointer to the next part of the list
  char string[120];        // string in list
};

/* --- globals --- */

int printnotdone;
char printername[120];
struct string_list *printernames = NULL;

/* --- function declarations --- */

void CreatePrintWindow(GtkWidget *widget);
void ListBoxAddItem(GtkWidget *listbox, char *sText, int printerdex);
void CloseWidget(GtkWidget *widget, gpointer data);
void PrintCallback (GtkWidget *widget, gpointer *data);
void ListItemSelectedCallback(GtkWidget *widget, gpointer *data);
struct string_list *string_list_append(struct string_list *list,char *string);
char *string_list_retrieve(struct string_list *list,int index);
void *string_list_destroy(struct string_list *list);

/* --- functions --- */

/* AddListBoxItem
 *
 * Adds an item to a list box
 */
void ListBoxAddItem(GtkWidget *listbox, char *sText,int printerdex)
{
  GtkWidget *item;

  item = gtk_list_item_new_with_label(sText);
  gtk_signal_connect(GTK_OBJECT(item),"select",
		     GTK_SIGNAL_FUNC(ListItemSelectedCallback),(int *)printerdex);
  gtk_container_add(GTK_CONTAINER(listbox),item);
}

/* CreatePrintWindow
 *
 * creates a window which should contain a list of avalible printers,
 * a print button, and a cancel button
 * 
 * shoulde be smart enough to identify if the widget passed to it
 * is a graph or a twodplot
 */
void CreatePrintWindow(GtkWidget *widget)
{
  char dummystr[120];
  char dummystr2[120];
  FILE *printcap;
  int test;
  GtkWidget *localwindow;
  GtkWidget *localbutton;
  GtkWidget *localvbox;
  GtkWidget *localhbox;
  GtkWidget *locallistbox;
  int i,j;

  /* --- idiot check --- */
  
  g_return_if_fail(widget != NULL);

  /* --- initalize the printernames GSList --- */
  /* Commented out continuous updating of printer names */
  //  if (printernames != NULL) {
  //  string_list_destroy(printernames);
  //  printernames = NULL;
  // }

  /* --- make the window --- */

  localwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(localwindow),"Gnuscope Print");
  gtk_window_set_modal(GTK_WINDOW(localwindow),TRUE);

  /* --- create the localvbox --- */
  localvbox = gtk_vbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(localwindow),localvbox);

  /* --- create the list box --- */
  locallistbox = gtk_list_new();
  gtk_list_set_selection_mode(GTK_LIST(locallistbox),GTK_SELECTION_SINGLE);
  gtk_container_add(GTK_CONTAINER(localvbox),locallistbox);

  /* --- open printcap and add the appropriate entries to the list box --- */
  i = 0;
  if ((printcap = fopen("/etc/printcap","r")) != NULL) {
    while (fgets(dummystr,120,printcap) != NULL) {
      if ((strncmp(dummystr," ",1) != 0) && 
	  (strncmp(dummystr,"#",1) != 0)) {
	//      if ((strncmp(dummystr,"l",1) == 0)) {
	j = 0;
	while (((strncmp(&dummystr[j],":",1) != 0) &&
		(strncmp(&dummystr[j],"|",1) != 0)) 
	       && (j < strlen(dummystr)))
	 j++;
	sprintf(dummystr2,"");
	strncat(dummystr2,dummystr,j);
	//sscanf(dummystr2,"%s",dummystr);
	printernames = string_list_append(printernames,dummystr2);
	ListBoxAddItem(locallistbox,dummystr2, i);
	i++;
      }
    }
    fclose(printcap);
  }

  /* --- create the localhbox for the print and cancel button --- */

  localhbox = gtk_hbox_new(FALSE,0);
  gtk_container_add(GTK_CONTAINER(localvbox),localhbox);

  /* --- create the printbutton --- */

  localbutton = gtk_button_new_with_label("Print");
  gtk_container_add(GTK_CONTAINER(localhbox),localbutton);
  gtk_signal_connect(GTK_OBJECT(localbutton),"clicked",
		     GTK_SIGNAL_FUNC(PrintCallback),(gpointer *)widget);
  gtk_signal_connect_object(GTK_OBJECT(localbutton),"clicked",
			    GTK_SIGNAL_FUNC(CloseWidget),GTK_OBJECT(localwindow));

  /* --- create the cancel button --- */

  localbutton = gtk_button_new_with_label("Cancel");
  gtk_container_add(GTK_CONTAINER(localhbox),localbutton);
  gtk_signal_connect_object(GTK_OBJECT(localbutton),"clicked",
			    GTK_SIGNAL_FUNC(CloseWidget),GTK_OBJECT(localwindow));

  /* --- show the window --- */
  gtk_widget_show_all(localwindow);
}

/* PrintCallback
 *
 * Prints based on what is selected in the listbox previously
 */
void PrintCallback (GtkWidget *widget, gpointer *data) 
{
  pid_t pid;
  char dummystr[120];

  sprintf(dummystr,"/tmp/gnuscope%d%d.ps",(int)time(NULL),getpid());
  if (GTK_IS_GRAPH(data))
    gtk_graph_make_ps(GTK_GRAPH(data),dummystr);
  if (GTK_IS_TWODPLOT(data))
    gtk_twodplot_make_ps(GTK_TWODPLOT(data),dummystr,0);
      
  while (printnotdone) {
    /* --- do nothing until done printing --- */
  }

  if ((pid = fork()) < 0)
    g_error("fork error");
  else if (pid == 0) {
    if ((pid = execl("/usr/bin/lpr","lpr","-r","-P",printername,dummystr,(char *) 0)) == -1)
      g_error("execl error");
  }
  //  if ((pid = fork()) < 0)
  //  g_error("fork error");
  //else if (pid == 0) {
  // if ((pid = execl("/bin/rm","rm",dummystr,(char *) 0)) == -1)
  //   g_error("execl error");
  //}
  if (waitpid(pid,NULL,0) < 0)
    g_error("wait error");
}

void ListItemSelectedCallback(GtkWidget *widget, gpointer *data)
{
  char dummystr[80];

  sprintf(dummystr,"%s",(char *) string_list_retrieve(printernames,(int) data));
  printf("%d , %s\n",(int) data,(char *) dummystr);
  sprintf(printername,"%s",(char *) dummystr);
}

struct string_list *string_list_append(struct string_list *list,char *string)
{
  struct string_list *temp;

  /* --- find the last element of the list --- */

  if (list != NULL) {
    temp = list;
    while (temp->ptr != NULL) {
      temp = temp->ptr;
    }
    /* --- at this point temp should point to the last element of the list --- */
    temp->ptr = (struct string_list *) malloc(sizeof(struct string_list));
    temp->ptr->ptr = NULL;
    sprintf(temp->ptr->string,"%s",string);
    return(list);
  } else {
    list = (struct string_list *) malloc(sizeof(struct string_list));
    list->ptr = NULL;
    sprintf(list->string,"%s",string);
  }
  return(list);
}

char *string_list_retrieve(struct string_list *list,int index)
{
  int counter;
  struct string_list *temp;

  temp = list;

  counter = 0;
  while ((counter != index) && (temp->ptr != NULL)) {
    temp = temp->ptr;
    counter ++;
  }
  return(temp->string);
}

/* string_list_destroy
 *
 * Attempts to destroy the singlely linked string list
 * Starting with the first element and working backwards
 */
void *string_list_destroy(struct string_list *list)
{
  struct string_list *temp;
  
  
  temp = list;

  while (list->ptr != NULL) {
    temp = list->ptr;
    free(list);
    list = NULL;
    list = temp->ptr;
  }
}
